import app.utilities.ssh as ssh


class PubKey:

    def __init__(self, label, handle, pem, private_key_handle):
        self.label = label
        self.handle = handle
        self.pem = pem
        self.private_key_handle = private_key_handle

    @classmethod
    def create(cls, ip_address, ssh_key_file_path, eni_ip, crypto_user_username, crypto_user_password, label):

        resp = ssh.gen_ecc_key_pair(
            ip_address=ip_address,
            ssh_key_file_path=ssh_key_file_path,
            eni_ip=eni_ip,
            crypto_user_username=crypto_user_username,
            crypto_user_password=crypto_user_password,
            key_label=label
        )
        if resp['status_code'] != 200:
            breakpoint()
        return cls(
            label=resp['data']['label'],
            pem=resp['data']['pem'],
            handle=resp['data']['handle'],
            private_key_handle=resp['data']['private_key_handle']
        )

    def read(self):
        return

    def update(self):
        return

    def delete(self):
        return



import requests
import json


def get_confirmed_sat_balance(address):
    r = requests.get(
        f"https://sochain.com/api/v2/get_address_balance/BTC/{address}", timeout=5)
    assert r.status_code == 200, f'API returned a status_code of {r.status_code}, please try again later'
    resp = r.json()['data']
    btc_bal = resp['confirmed_balance']
    sat_bal = int(float(btc_bal) * 100000000)

    return sat_bal


def get_tx_inputs(address):
    resp = requests.get(
        f"https://sochain.com//api/v2/get_tx_unspent/BTC/{address}", timeout=5)
    assert resp.status_code == 200, f'block explorer return status code {resp.status_code}'
    resp_tx_inputs = resp.json()['data']['txs']
    assert not not resp_tx_inputs, f"Address: {address} has no unspent transactions"
    tx_inputs = []
    for resp_tx in resp_tx_inputs:
        tx = {}
        tx['output_no'] = resp_tx['output_no']
        tx['outpoint_index'] = (resp_tx['output_no']).to_bytes(
            4, byteorder='little', signed=False)
        hash = bytearray.fromhex(resp_tx['txid'])
        hash.reverse()
        tx['outpoint_hash'] = hash
        tx_inputs.append(tx)

    return tx_inputs



from app.adapters import Explorer

import hashlib
from base58 import b58encode_check
from ecdsa import VerifyingKey


class P2PKH:
    def __init__(self, pem):
        self.public_key = pem

    @property
    def btc_public_key(self):
        vk = VerifyingKey.from_pem(self.public_key)
        btc_public_key = bytes.fromhex('04') + vk.to_string()

        return btc_public_key.hex()

    @property
    def address(self):
        sha256 = hashlib.sha256(bytes.fromhex(self.btc_public_key)).digest()

        ripemd160_hash = hashlib.new('ripemd160')
        ripemd160_hash.update(sha256)
        ripemd160 = ripemd160_hash.digest()

        hashed_btc_public_key = bytes.fromhex('00') + ripemd160

        addr = b58encode_check(hashed_btc_public_key).decode('utf-8')

        return addr

    @property
    def confirmed_balance(self):
        explorer = Explorer(address=self.address)
        return explorer.confirmed_balance

    @property
    def spent(self):
        explorer = Explorer(address=self.address)
        return explorer.spent



import app.adapters.blockcypher_adapters as blockcypher

import datetime


class Explorer:
    def __init__(self, address):
        self.address = address
        address_data = _get_address_data(address=address)
        self.confirmed_balance = address_data['confirmed_balance']
        self.spent = address_data['spent']
        self.txrefs = _seralize_datetime(txrefs=address_data['txrefs'])


def _get_address_data(address):
    resp = blockcypher.address_data(address=address)

    return {
        'address': resp['address'],
        'confirmed_balance': resp['confirmed_balance'],
        'spent': resp['spent'],
        'txrefs': resp['txrefs']
    }


def _seralize_datetime(txrefs):
    for txref in txrefs:
        for k, v in txref.items():
            if isinstance(v, datetime.datetime):
                txref[k] = v.__str__()

    return txrefs


# def blockcypher_address_tx_inputs(txrefs):
#     tx_inputs = []
#     if bool(txrefs):
#         for txref in txrefs:
#             tx = {}
#             tx['output_no'] = txref['tx_output_n']
#             if tx['output_no'] < 0:
#                 tx['outpoint_index'] = ''
#                 tx['outpoint_hash'] = ''
#             else:
#                 tx['outpoint_index'] = (txref['tx_output_n']).to_bytes(
#                     4, byteorder='little', signed=False)
#                 hash = bytearray.fromhex(txref['tx_hash'])
#                 hash.reverse()
#                 tx['outpoint_hash'] = hash
#             tx_inputs.append(tx)
#     return tx_inputs



from app.controllers.credentials_controller import CredentialsController
from app.models.pub_key_model import PubKey

from app.models.instance_model import Instance
from app.models.cluster_model import Cluster
from app.models.address_model import Address

import uuid
import os


class AddressController:

    def __init__(self, credentials):
        self.credentials = credentials

    def index(self):
        addresses = Address.all(bucket=self.bucket_name,
                                s3=self.credentials.s3)
        for address in addresses:
            self.update(id=address.id)

        addresses = Address.all(bucket=self.bucket_name,
                                s3=self.credentials.s3)
        return {'data': {'addresses': addresses}, 'http_status_code': 200}

    def create(self):
        pub_key = PubKey.create(
            ip_address=self.ip_address,
            ssh_key_file_path=self.ssh_key_file_path,
            eni_ip=self.eni_ip,
            crypto_user_username=self.credentials.data['crypto_user_username'],
            crypto_user_password=self.credentials.data['crypto_user_password'],
            label=_get_address_id()
        )
        address = Address.create(pub_key=pub_key)
        assert address.save(
            bucket=self.bucket_name,
            s3=self.credentials.s3,
            region=self.credentials.data['aws_region']
        ), f"Failed to save address: {address.id} to bucket: {self.bucket_name}"
        return {'data': {'address': address}, 'http_status_code': 200}

    def show(self, id):
        address = Address.find(bucket=self.bucket_name,
                               s3=self.credentials.s3, id=id)
        return {'data': {'address': address}, 'http_status_code': 200}

    def update(self, id):
        address = Address.find(
            id=id,
            bucket=self.bucket_name,
            s3=self.credentials.s3
        ).refresh(
            bucket=self.bucket_name,
            s3=self.credentials.s3,
            region=self.credentials.data['aws_region']
        )
        return {'data': {'address': address}, 'http_status_code': 200}

    @property
    def ip_address(self):
        instance_id = self.credentials.data['instance_id']
        instance = Instance(id=instance_id, resource=self.credentials.resource)
        ip_address = instance.public_ip_address
        return ip_address

    @property
    def ssh_key_file_path(self):
        ssh_key_file_path = os.path.join(
            self.credentials.data['cluster_path'],
            f"{self.credentials.data['ssh_key_name']}.pem"
        )
        return ssh_key_file_path

    @property
    def eni_ip(self):
        cluster = Cluster(
            id=self.credentials.data['cluster_id'], client=self.credentials.cloudhsmv2)
        hsms = cluster.hsms
        if bool(hsms) is False:
            raise NoHSMFoundError(f'Cluster: {cluster.id} has no HSMs')
        eni_ip = hsms[0]['EniIp']
        return eni_ip

    @property
    def bucket_name(self):
        bucket_name = f"{self.credentials.data['cluster_id']}-bucket"
        return bucket_name


def _get_address_id():
    return f'addr-{str(uuid.uuid4())[-12:]}'


class NoHSMFoundError(Exception):
    pass


class SSHKeyFileNotFoundError(Exception):
    pass



import app.utilities.ssh as ssh


class Instance:

    def __init__(self, resource, id):
        self.resource = resource
        self.id = id

    @classmethod
    def all(cls, client):
        instances = []
        resp = client.describe_instances()
        for reservation in resp['Reservations']:
            instance = reservation['Instances'][0]
            if instance['State']['Name'] != 'terminated':
                instances.append(instance)
        return instances

    @property
    def public_ip_address(self):
        instance = self.resource.Instance(self.id)
        return instance.public_ip_address

    @property
    def state(self):
        instance = self.resource.Instance(self.id)
        return instance.state['Name']

    @property
    def ssh_key_name(self):
        instance = self.resource.Instance(self.id)
        return instance.key_name

    def install_packages(self, ssh_key_file_path):
        outputs = ssh.install_packages(
            ip_address=self.public_ip_address, ssh_key_file_path=ssh_key_file_path)
        return True

    def start(self):
        instance = self.resource.Instance(self.id)
        resp = instance.start()
        return resp

    def stop(self):
        instance = self.resource.Instance(self.id)
        resp = instance.stop()
        return resp




from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, padding
import datetime


class Certs:

    def __init__(self, pem_csr, passphrase):
        self.pem_csr = pem_csr
        self.passphrase = passphrase

        self._csr = csr(self.pem_csr)
        self._key = private_key()
        self._ca_cert = ca_cert(self._key)
        self._hsm_cert = hsm_cert(self._ca_cert, self._csr, self._key)

        self.valid = validate_hsm_cert(self._hsm_cert, self._ca_cert)
        self.pem_private_key = pem_private_key(self._key, self.passphrase)
        self.pem_ca_cert = pem_ca_cert(self._ca_cert)
        self.pem_hsm_cert = pem_hsm_cert(self._hsm_cert)


def csr(pem_csr):
    if type(pem_csr) is bytes:
        bytes_csr = pem_csr
    else:
        bytes_csr = bytes(pem_csr, 'utf-8')

    return x509.load_pem_x509_csr(bytes_csr)


def private_key():
    return rsa.generate_private_key(public_exponent=65537, key_size=2048)


def private_key_from_pem(pem_private_key, string_passphrase):
    return serialization.load_pem_private_key(pem_private_key, bytes(string_passphrase, 'UTF-8'))


def pem_private_key(bytes_private_key, string_passphrase):
    pem_sk = bytes_private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.BestAvailableEncryption(
            bytes(string_passphrase, 'utf-8')))
    return pem_sk


def ca_cert(private_key):
    builder = x509.CertificateBuilder()
    name = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"."),
        x509.NameAttribute(NameOID.LOCALITY_NAME, u"."),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"."),
        x509.NameAttribute(NameOID.COMMON_NAME, u".")
    ])

    builder = builder.serial_number(1)
    builder = builder.issuer_name(name)
    builder = builder.subject_name(name)
    builder = builder.public_key(private_key.public_key())
    builder = builder.not_valid_before(
        datetime.datetime.utcnow() - datetime.timedelta(hours=1))
    builder = builder.not_valid_after(
        datetime.datetime.utcnow() + datetime.timedelta(days=3652))
    builder = builder.add_extension(x509.BasicConstraints(ca=True, path_length=None), True
                                    )
    return builder.sign(private_key, hashes.SHA256(), default_backend())


def pem_ca_cert(ca_cert):
    return ca_cert.public_bytes(serialization.Encoding.PEM)


def hsm_cert(ca_cert, csr, private_key):
    builder = x509.CertificateBuilder(
        issuer_name=ca_cert.issuer,
        subject_name=csr.subject,
        public_key=csr.public_key(),
        not_valid_before=datetime.datetime.utcnow() - datetime.timedelta(hours=1),
        not_valid_after=datetime.datetime.utcnow() + datetime.timedelta(days=3652),
        extensions=csr.extensions,
        serial_number=x509.random_serial_number()
    )
    return builder.sign(
        private_key=private_key, algorithm=hashes.SHA256(), backend=default_backend())


def pem_hsm_cert(hsm_cert):
    return hsm_cert.public_bytes(serialization.Encoding.PEM)


def validate_hsm_cert(hsm_cert, ca_cert):
    public_key = ca_cert.public_key()

    try:
        public_key.verify(
            hsm_cert.signature,
            hsm_cert.tbs_certificate_bytes,
            padding.PKCS1v15(),
            hsm_cert.signature_hash_algorithm
        )
        return True
    except:
        return False



from base58 import b58decode_check
import hashlib
from ecdsa import SECP256k1, VerifyingKey, util


class TxOutputScript():

    def __init__(self, address, confirmed_balance, recipient, fee, value, change_address):
        assert address != change_address, 'An address cannot send change to itself'
        self.confirmed_balance = confirmed_balance
        self.recipient = recipient
        self.fee = fee
        self.value = value
        self.change_address = change_address

    @property
    def outputs(self):
        assert self.fee > 0, 'fee must be greater than zero'
        assert self.verify_address(
            self.recipient), 'reipient address is not valid'
        if self.value is None:
            self.value = self.confirmed_balance - self.fee
        else:
            assert self.value > 0, 'quantity must be greater than zero'
            assert self.confirmed_balance >= self.value + \
                self.fee, 'quanity + fee cannot be greater than confirmed balance'

        outputs = []
        outputs.append(self.output(self.value, self.recipient))

        if self.value + self.fee < self.confirmed_balance:
            assert self.verify_address(
                self.change_address), 'change address is not valid'
            change_value = self.confirmed_balance - self.value - self.fee
            outputs.append(self.output(change_value, self.change_address))

        return outputs

    def verify_address(self, address):
        # TO DO
        return True

    def output(self, value, address):
        pub_key_script = PubKeyScript(address)
        output = bytearray()
        output += (value).to_bytes(8, byteorder="little", signed=False)
        output += pub_key_script.bytes
        output += pub_key_script.script

        return output


class PubKeyScript():

    def __init__(self, address):
        self.address = address
        pass

    @property
    def script(self):
        if self.address[0] == '1':
            return self.p2pkh_pk_script
        elif self.address[0] == '3':
            return self.p2sh_pk_script
        else:
            raise ('Only P2PKH and P2SH addresses are supported')

    @property
    def bytes(self):
        return len(self.script).to_bytes(
            1, byteorder='little', signed=False)

    @property
    def addressToPubKeyHash(self):
        return b58decode_check(self.address)[1:]

    @property
    def p2pkh_pk_script(self):
        script = bytearray()
        script += bytearray.fromhex('76')  # OP_DUP
        script += bytearray.fromhex('a9')  # OP_HASH
        script += bytearray.fromhex('14')  # Push 20 bytes as Data
        script += self.addressToPubKeyHash
        script += bytearray.fromhex('88')  # OP_EQUALVERIFY
        script += bytearray.fromhex('ac')  # OP_CHECKSIG
        return script

    @property
    def p2sh_pk_script(self):
        script = bytearray()
        script += bytearray.fromhex('a9')  # OP_HASH
        script += bytearray.fromhex('14')  # Push 20 bytes as Data
        script += self.addressToPubKeyHash
        script += bytearray.fromhex('87')  # OP_EQUAL
        return script


class SigScript():

    def __init__(self, signatures, unsigned_tx, tx_in, pem):
        self.signatures = signatures
        self.unsigned_tx = unsigned_tx
        self.tx_in = tx_in
        self.pem = pem

    @property
    def signature(self):
        for msg in self.unsigned_tx.messages:
            if msg['tx_input'] == self.tx_in:
                message = msg['message']

        for signature in self.signatures:

            if self.verify_der_signature(signature, message, self.pem):
                return self.der_canonize(signature)

        raise SingatureNotValid('Signatures are not valid')

    @property
    def signature_bytes(self):
        return (len(self.signature) + 1).to_bytes(1,
                                                  byteorder='little',
                                                  signed=False)

    @property
    def public_key(self):
        vk = VerifyingKey.from_pem(self.pem)
        return bytes.fromhex('04') + vk.to_string()

    @property
    def public_key_bytes(self):
        return (len(self.public_key)).to_bytes(1,
                                               byteorder="little",
                                               signed=False)

    @property
    def script(self):
        sig_script = bytearray()
        sig_script += self.signature_bytes
        sig_script += self.signature
        sig_script += bytes.fromhex('01')
        sig_script += self.public_key_bytes
        sig_script += self.public_key
        return sig_script

    @property
    def script_bytes(self):
        return (len(self.script)).to_bytes(1,
                                           byteorder='little',
                                           signed=False)

    def verify_der_signature(self, signature, message, pem):
        double_hashed_message = hashlib.sha256(
            hashlib.sha256(message).digest()).digest()

        try:
            vk = VerifyingKey.from_pem(pem)
            vk.verify_digest(signature, double_hashed_message,
                             sigdecode=util.sigdecode_der)
            return True
        except Exception:
            return False

    def der_canonize(self, signature):
        r, s = util.sigdecode_der(signature, SECP256k1.order)
        return util.sigencode_der_canonize(r, s, SECP256k1.order)


class SingatureNotValid(Exception):
    pass



from app.utilities.bitcoin.tx_scripts import SigScript
from app.utilities import ssh

import os


class SignedTx:

    def __init__(self, pem, unsigned_tx, signatures):
        self.pem = pem
        self.unsigned_tx = unsigned_tx
        self.signatures = signatures

    @classmethod
    def create(cls, unsigned_tx, address, credentials, ip_address, eni_ip):
        unsigned_tx_files = _unsigned_tx_files(
            unsigned_tx=unsigned_tx, cluster_path=credentials.data['cluster_path'])

        ssh_key_file_path = os.path.join(
            credentials.data['cluster_path'], f"{credentials.data['ssh_key_name']}.pem")

        signatures = _get_signatures(
            unsigned_tx_files=unsigned_tx_files,
            ssh_key_file_path=ssh_key_file_path,
            eni_ip=eni_ip,
            ip_address=ip_address,
            pub_key_handle=address.pub_key_handle,
            private_key_handle=address.private_key_handle,
            crypto_user_username=credentials.data['crypto_user_username'],
            crypto_user_password=credentials.data['crypto_user_password'],
            cluster_path=credentials.data['cluster_path']
        )

        return cls(
            pem=address.pub_key_pem,
            unsigned_tx=unsigned_tx,
            signatures=signatures
        )

    @property
    def hex(self):
        tx = bytearray()

        tx += self.unsigned_tx.version
        tx += self.unsigned_tx.tx_in_count
        for tx_in in self.unsigned_tx.tx_inputs:

            sig_script = SigScript(
                self.signatures, self.unsigned_tx, tx_in, self.pem)

            tx += self.unsigned_tx.previous_output(tx_in)
            tx += sig_script.script_bytes
            tx += sig_script.script
            tx += self.unsigned_tx.sequence
        tx += self.unsigned_tx.tx_out_count
        for tx_out in self.unsigned_tx.tx_outs.outputs:
            tx += tx_out
        tx += self.unsigned_tx.lock_time

        return tx.hex()


def _unsigned_tx_files(unsigned_tx, cluster_path):
    n, to_sign_files = 0, []
    for sign in unsigned_tx.to_sign:
        n += 1
        to_sign_file_path = os.path.join(cluster_path, f'unsigned_tx_{n}.bin')
        with open(to_sign_file_path, 'wb') as file:
            file.write(sign)
            to_sign_files.append(
                {'file_path': to_sign_file_path, 'file_name': f'unsigned_tx_{n}.bin'})

    return to_sign_files


def _get_signatures(unsigned_tx_files, ssh_key_file_path, eni_ip, ip_address, pub_key_handle,
                    private_key_handle, crypto_user_username, crypto_user_password, cluster_path):
    count = 0
    signatures = []
    for unsigned_tx_file in unsigned_tx_files:
        count += 1
        signature_file_path = ssh.sign_tx(
            ip_address=ip_address,
            ssh_key_file_path=ssh_key_file_path,
            eni_ip=eni_ip,
            unsigned_tx_file=unsigned_tx_file,
            pub_key_handle=pub_key_handle,
            private_key_handle=private_key_handle,
            crypto_user_username=crypto_user_username,
            crypto_user_password=crypto_user_password,
            count=count,
            cluster_path=cluster_path
        )

        with open(signature_file_path, 'rb') as file:
            signature = file.read()
        signatures.append(signature)
        os.remove(unsigned_tx_file['file_path'])
        os.remove(signature_file_path)

    return signatures



from app.controllers.credentials_controller import CredentialsController
from app.controllers.addresses_controller import AddressController

from app.models.instance_model import Instance
from app.models.cluster_model import Cluster
from app.models.unsigned_tx_model import UnsignedTx
from app.models.signed_tx_model import SignedTx

from app.adapters import Explorer
import boto3
from botocore.exceptions import ClientError

import os


class TxController:

    def __init__(self, credentials):

        self.credentials = credentials

    def create(self, address, recipient, fee, value, change_address=None, change=None, balance=None):
        unsigned_tx = UnsignedTx(
            address=address,
            recipient=recipient,
            fee=fee,
            value=value,
            change_address=change_address
        )

        instance = Instance(resource=self.credentials.resource,
                            id=self.credentials.data['instance_id'])

        cluster = Cluster(client=self.credentials.cloudhsmv2,
                          id=self.credentials.data['cluster_id'])
        eni_ip = cluster.hsms[0]['EniIp']

        signed_tx = SignedTx.create(
            unsigned_tx=unsigned_tx,
            address=address,
            credentials=self.credentials,
            ip_address=instance.public_ip_address,
            eni_ip=eni_ip
        )

        return signed_tx.hex

    def validate(self, all, address_id, recipient, fee, value, change_address):
        try:
            controller = AddressController(credentials=self.credentials)
            resp = controller.update(id=address_id)
            address = resp['data']['address']
            confirmed_balance = address.confirmed_balance
            assert confirmed_balance > 0, f"Address {address.address} has a zero confirmed balance."
            assert fee > 0, f"Mining fee is zero."

            if all:
                assert fee < confirmed_balance, f"The mining fee of {fee} is greater than the confirmed balance of {confirmed_balance} in address {address.address}."
                assert _is_address_valid(
                    address=recipient),  f"Recipient address {recipient} does not appear to be valid."
                return {
                    'address': address,
                    'balance': confirmed_balance,
                    'recipient': recipient,
                    'value': confirmed_balance - fee,
                    'fee': fee
                }

            else:
                assert fee + \
                    value < confirmed_balance, f"The mining fee of {fee} and the quantiy of {value} are greater than the confirmed balance of {confirmed_balance} in address {address.address}."
                assert _is_address_valid(
                    address=recipient), f"Recipient address {recipient} does not appear to be valid."
                assert _is_address_valid(
                    address=change_address), f"Change address {change_address} does not appear to be valid."
                return {
                    'address': address,
                    'balance': confirmed_balance,
                    'recipient': recipient,
                    'value': value,
                    'fee': fee,
                    'change': confirmed_balance - fee - value,
                    'change_address': change_address
                }

        except ClientError:
            return {'error': f"Address ID {address_id} was not found"}

        except Exception as e:
            return {'error': e.args[0]}


def _unsigned_tx_files(unsigned_tx, path):
    n, to_sign_files = 0, []
    for sign in unsigned_tx.to_sign:
        n += 1
        to_sign_file = os.path.join(path, f'unsigned_tx_{n}.bin')
        with open(to_sign_file, 'wb') as file:
            file.write(sign)
            to_sign_files.append(to_sign_file)

    return to_sign_files


def _is_address_valid(address):
    explorer = Explorer(address=address)
    try:
        resp = explorer.spent
        return True
    except Exception as e:
        return False



import blockcypher


def address_data(address):
    try:
        resp = blockcypher.get_address_details(address=address)

        return {
            'address': resp['address'],
            'confirmed_balance': resp['balance'],
            'txrefs': resp.get('txrefs'),
            'spent': is_spent(data=resp)
        }

    except AssertionError as e:
        raise AddressNotValid(e.args[0])

    except Exception as e:
        raise Exception(e.args[0])


def is_spent(data):
    if data['total_sent'] > 0:
        return True
    else:
        return False


class AddressNotValid(Exception):
    pass



from app.models.cluster_model import Cluster


class HSM():

    def __init__(self, id, cluster_id, client):
        self.id = id
        self.cluster_id = cluster_id
        self.client = client

    @classmethod
    def create(cls, cluster_id, availability_zone, client):
        resp = client.create_hsm(ClusterId=cluster_id,
                                 AvailabilityZone=availability_zone)
        hsm_id = resp['Hsm']['HsmId']
        return cls(id=hsm_id, cluster_id=cluster_id, client=client)

    @property
    def state(self):
        return self.read()['State']

    def read(self):
        resp = self.client.describe_clusters(
            Filters={'clusterIds': [self.cluster_id]})
        hsms = resp['Clusters'][0]['Hsms']
        for hsm in hsms:
            if hsm['HsmId'] == self.id:
                return hsm
        return False

    def destroy(self):
        resp = self.client.delete_hsm(
            ClusterId=self.cluster_id, HsmId=self.id)
        return resp['HsmId']



from app.models.cluster_model import Cluster
from app.models.hsm_model import HSM
from app.models.instance_model import Instance
import json


class StatusController:
    def __init__(self, credentials):
        self.credentials = credentials
        self.base_path = credentials.data['base_path']
        self.ec2 = credentials.ec2
        self.resource = credentials.resource
        self.cloudhsmv2 = credentials.cloudhsmv2

    def show(self):
        # hsm states = ['CREATE_IN_PROGRESS', 'ACTIVE', 'DELETE_IN_PROGRESS']
        # instance states = ['pending', 'running', 'stopping', 'stopped' ]

        hsm_state = self.hsm.state if bool(self.hsm) else None
        instance_state = self.instance.state

        if hsm_state == 'CREATE_IN_PROGRESS' and instance_state == 'pending':
            status = 'Piggy is waking up'
        elif hsm_state == 'CREATE_IN_PROGRESS' and instance_state == 'running':
            status = 'Piggy is waking up'
        elif hsm_state == 'ACTIVE' and instance_state == 'pending':
            status = 'Piggy is waking up'

        elif hsm_state == 'ACTIVE' and instance_state == 'running':
            status = 'Piggy is awake'

        elif hsm_state == 'DELETE_IN_PROGRESS' and instance_state == 'stopping':
            status = 'Piggy is going to sleep'
        elif hsm_state == 'DELETE_IN_PROGRESS' and instance_state == 'stopped':
            status = 'Piggy is going to sleep'
        elif hsm_state is None and instance_state == 'stopping':
            status = 'Piggy is going to sleep'

        elif hsm_state is None and instance_state == 'stopped':
            status = 'Piggy is sleeping'

        else:
            status = 'groogy, please run piggy status -wake or -sleep'

        return status

    def sleep(self):
        status = self.show()
        if status == 'Piggy is sleeping':
            return 'Piggy is sleeping'
        elif status == 'Piggy is going to sleep':
            return 'Piggy is going to sleep'
        elif status == 'Piggy is waking up':
            return 'Piggy is waking up, please try agian once Piggy is awake'
        else:
            hsm_resp = self.hsm.destroy()
            instance_resp = self.instance.stop()
        return 'Piggy is going to sleep'

    def wake(self):
        status = self.show()
        if status == 'Piggy is awake':
            return 'Piggy is awake'
        elif status == 'Piggy is waking up':
            return 'Piggy is waking up'
        elif status == 'Piggy is going to sleep':
            return 'Piggy is going to sleep, please try again once Piggy is sleeping'
        else:
            cluster = self.cluster
            hsm_resp = HSM.create(
                cluster_id=cluster.id,
                availability_zone=cluster.azs[0],
                client=self.cloudhsmv2
            )
            instance_resp = self.instance.start()
        return 'Piggy is waking up'

    @property
    def instance(self):
        id = self.credentials.data['instance_id']
        instances = Instance.all(client=self.ec2)
        if len(instances) == 0:
            raise InstanceNotFoundError(f'Instance ID {id} not found.')
        elif len(instances) > 1:
            raise MultipleInstancesFoundError('Multiple instances found.')
        elif instances[0]['InstanceId'] != id:
            raise InstanceNotFoundError(f'Instance ID {id} not found.')
        else:
            instance = Instance(id=id, resource=self.resource)
            return instance

    @property
    def cluster(self):
        id = self.credentials.data['cluster_id']
        clusters = Cluster.all(client=self.cloudhsmv2)
        if len(clusters) == 0:
            raise ClusterNotFoundError(f'Cluster ID {id} not found.')
        elif len(clusters) > 1:
            raise MultipleClustersFoundError('Multiple clusters found.')
        elif clusters[0]['ClusterId'] != id:
            raise ClusterNotFoundError(f'Cluster ID {id} not found.')
        else:
            cluster = Cluster(id=id, client=self.cloudhsmv2)
            if cluster.state != 'ACTIVE':
                raise ClusterStateError('Cluster state must be ACTIVE.')
            return cluster

    @property
    def hsm(self):
        hsms = self.cluster.hsms
        if len(hsms) == 0:
            return None
        elif len(hsms) > 1:
            raise MultipleHSMsFoundError('Multiple HSMs found.')
        else:
            id = hsms[0]['HsmId']
            hsm = HSM(id=id, cluster_id=self.cluster.id,
                      client=self.cloudhsmv2)
            return hsm

    def backups(self):
        backups = []
        resp = self.cloudhsmv2.describe_backups()
        for backup in resp['Backups']:
            backups.append(
                {
                    'id': backup['BackupId'], 'state': backup['BackupState'],
                    'created': backup['CreateTimestamp'].strftime('%m/%d/%y %H:%M:%S')
                }
            )
        return backups


class InstanceNotFoundError(Exception):
    pass


class MultipleInstancesFoundError(Exception):
    pass


class ClusterNotFoundError(Exception):
    pass


class MultipleClustersFoundError(Exception):
    pass


class ClusterStateError(Exception):
    pass


class HSMNotFoundError(Exception):
    pass


class MultipleHSMsFoundError(Exception):
    pass



import json
import os
import boto3


class CredentialsData:
    def __init__(self, **kwargs):
        self.base_path = kwargs['base_path']
        self.cluster_path = kwargs['cluster_path']
        self.aws_region = kwargs['aws_region']
        self.ssh_key_name = kwargs['ssh_key_name']
        self.cluster_id = kwargs['cluster_id']
        self.aws_access_key_id = kwargs['aws_access_key_id']
        self.aws_secret_access_key = kwargs['aws_secret_access_key']
        self.customer_ca_key_password = kwargs['customer_ca_key_password']
        self.crypto_officer_password = kwargs['crypto_officer_password']
        self.crypto_user_username = kwargs['crypto_user_username']
        self.crypto_user_password = kwargs['crypto_user_password']
        self.instance_id = kwargs['instance_id']


def get_credentials_data(**kwargs):
    credentials_data = CredentialsData(**kwargs)
    return credentials_data.__dict__


class Credentials:
    def __init__(self, credentials_file_path, data):
        self.credentials_file_path = credentials_file_path
        self.data = data
        self.ec2 = boto3.client(
            'ec2',
            aws_access_key_id=self.data['aws_access_key_id'],
            aws_secret_access_key=self.data['aws_secret_access_key'])
        self.cloudhsmv2 = boto3.client('cloudhsmv2',
                                       aws_access_key_id=self.data['aws_access_key_id'],
                                       aws_secret_access_key=self.data['aws_secret_access_key'])
        self.resource = boto3.resource('ec2',
                                       aws_access_key_id=self.data['aws_access_key_id'],
                                       aws_secret_access_key=self.data['aws_secret_access_key'])
        self.s3 = boto3.client('s3',
                               aws_access_key_id=self.data['aws_access_key_id'],
                               aws_secret_access_key=self.data['aws_secret_access_key'])

    @classmethod
    def create(cls, credentials_file_path, data):

        _write_credentials_to_file(
            credentials_file_path=credentials_file_path, data=data)

        credentials = Credentials(
            credentials_file_path=credentials_file_path, data=data)

        return credentials

    @classmethod
    def read(cls, credentials_file_path):
        with open(credentials_file_path, 'r') as file:
            json_file_data = file.read()
        file_data = json.loads(json_file_data)
        data = get_credentials_data(**file_data)
        credentials = Credentials(
            credentials_file_path=credentials_file_path, data=data)

        return credentials

    def update(self, **kwargs):
        credentials_data = CredentialsData(**self.data)
        for key, value in kwargs.items():
            setattr(credentials_data, key, value)
        self.data = credentials_data.__dict__
        _write_credentials_to_file(
            credentials_file_path=self.credentials_file_path, data=self.data)
        return self


def _write_credentials_to_file(credentials_file_path, data):
    if os.path.exists(data['cluster_path']) is False:
        os.mkdir(data['cluster_path'])
    with open(credentials_file_path, 'w') as file:
        file.write(json.dumps(data))

    return



from functools import wraps
from app.controllers.status_controller import StatusController
from app.controllers.credentials_controller import CredentialsController
import boto3
import os
import click
from string import ascii_uppercase


def check_status(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        controller = StatusController(credentials=kwargs['credentials'])

        status = controller.show()
        if status == 'Piggy is sleeping':
            click.echo()
            click.secho(
                'Please run piggy status -wake before running this command', fg='red')
            click.echo()
            quit()
        elif status == 'Piggy is waking up':
            click.echo()
            click.secho(
                'Please wait unitl piggy status is Active before running this command', fg='red')
            click.echo()
            quit()
        else:
            pass

        return f(*args, **kwargs)
    return wrapper


def creds(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if kwargs.get('credentials_file_path') is None:

            path = '/Volumes/Piggy'
            if os.path.isdir(path) is False:
                creds_not_found()
            else:
                dirs = [d for d in os.listdir(
                    path) if os.path.isdir(os.path.join(path, d))]
                clusters = [c for c in dirs if 'cluster' in c]
                if len(clusters) == 0:
                    creds_not_found()
                elif len(clusters) > 1:
                    cluster = which_cluster(clusters=clusters)
                else:
                    cluster = clusters[0]

                credentials_file_path = os.path.join(
                    path, cluster, 'credentials.json')
        else:
            credentials_file_path = kwargs['credentials_file_path']

        if os.path.exists(credentials_file_path) is False:
            creds_not_found()

        try:
            controller = CredentialsController()
            credentials = controller.create_from_file(
                credentials_file_path=credentials_file_path)

        except Exception as e:
            creds_not_found()

        del kwargs['credentials_file_path']
        kwargs['credentials'] = credentials

        return f(*args, **kwargs)
    return wrapper


def creds_not_found():
    click.echo()
    click.secho(
        'Piggy could not find or your credentials.json file is not valid.', fg='red')
    click.echo()
    click.secho(
        'Please check your credentials.json file and run command with the -creds option.', fg='red')
    click.echo()
    quit()


def which_cluster(clusters):
    click.echo()
    click.secho('Piggy found multiple clusters.', fg='green')
    click.echo()
    choices = []
    for index, cluster in enumerate(clusters):
        click.secho(f'{ascii_uppercase[index]} - {cluster}', fg='green')
        choices.append(index+1)
    click.echo()
    choice = click.prompt('Please select a cluster:',
                          type=click.Choice(ascii_uppercase[:len(clusters)], case_sensitive=False))
    i = ascii_uppercase.index(choice)
    cluster = clusters[i]
    return cluster


class PiggyNotActive(Exception):
    pass



from app.utilities.bitcoin.tx_scripts import TxOutputScript, PubKeyScript
from app.utilities.bitcoin.addresses import P2PKH
from app.adapters import Explorer

import hashlib


class UnsignedTx:

    def __init__(self, address, recipient, fee, value, change_address):
        self.confirmed_balance = address.confirmed_balance
        self.tx_outs = TxOutputScript(
            address=address.address,
            confirmed_balance=address.confirmed_balance,
            recipient=recipient,
            fee=fee,
            value=value,
            change_address=change_address
        )
        self.pub_key_script = PubKeyScript(address=address.address)
        self.txrefs = address.txrefs

    @property
    def messages(self):
        messages = []
        for tx in self.tx_inputs:
            msg = bytearray()
            msg += self.version
            msg += self.tx_in_count
            for tx_in in self.tx_inputs:
                msg += self.previous_output(tx_in)
                if tx == tx_in:
                    msg += self.pub_key_script.bytes
                    msg += self.pub_key_script.script
                else:
                    msg += self.placeholder
                msg += self.sequence
            msg += self.tx_out_count
            for tx_output in self.tx_outs.outputs:
                msg += tx_output
            msg += self.lock_time
            msg += self.hash_code

            messages.append({'message': msg, 'tx_input': tx})

        return messages

    @property
    def to_sign(self):
        to_signs = []
        for msg in self.messages:
            to_signs.append(hashlib.sha256(msg['message']).digest())
        return to_signs

    @property
    def version(self):
        return (1).to_bytes(4, byteorder="little", signed=True)

    @property
    def tx_inputs(self):
        tx_inputs = []
        if bool(self.txrefs):
            for txref in self.txrefs:
                tx_input = {'output_no': txref['tx_output_n'],
                            'outpoint_index': None, 'outpoint_hash': None}
                if tx_input['output_no'] >= 0:
                    tx_input['outpoint_index'] = (txref['tx_output_n']).to_bytes(
                        4, byteorder='little', signed=False)
                    hash = bytearray.fromhex(txref['tx_hash'])
                    hash.reverse()
                    tx_input['outpoint_hash'] = hash
                tx_inputs.append(tx_input)
        return tx_inputs

    @property
    def tx_in_count(self):
        return (len(self.tx_inputs)).to_bytes(1, byteorder="little", signed=False)

    def previous_output(self, tx_in):
        output = bytearray()
        output += tx_in['outpoint_hash']
        output += tx_in['outpoint_index']
        return output

    @property
    def placeholder(self):
        return (0).to_bytes(1, byteorder='little', signed=False)

    @property
    def sequence(self):
        return bytes.fromhex("ffffffff")

    @property
    def tx_out_count(self):
        return (len(self.tx_outs.outputs)).to_bytes(1, byteorder="little", signed=False)

    @property
    def lock_time(self):
        return (0).to_bytes(4, byteorder="little", signed=False)

    @property
    def hash_code(self):
        return (1).to_bytes(4, byteorder='little', signed=False)



from app.models.credentials_model import Credentials
import json

import os


class CredentialsController:

    def __init__(self):
        pass

    def create(self, base_path, cluster_path, aws_region, aws_access_key_id, aws_secret_access_key, customer_ca_key_password,
               crypto_officer_password, crypto_user_username, crypto_user_password, cluster_id, instance_id, ssh_key_name):
        credentials_file_path = os.path.join(
            cluster_path, 'credentials.json')
        data = {
            'base_path': base_path,
            'cluster_path': cluster_path,
            'aws_region': aws_region,
            'aws_access_key_id': aws_access_key_id,
            'aws_secret_access_key': aws_secret_access_key,
            'customer_ca_key_password': customer_ca_key_password,
            'crypto_officer_password': crypto_officer_password,
            'crypto_user_username': crypto_user_username,
            'crypto_user_password': crypto_user_password,
            'cluster_id': cluster_id,
            'instance_id': instance_id,
            'ssh_key_name': ssh_key_name
        }

        credentials = Credentials.create(
            credentials_file_path=credentials_file_path, data=data)
        return credentials

    def create_from_file(self, credentials_file_path):
        return Credentials.read(credentials_file_path=credentials_file_path)

    def show(self, credentials_file_path):
        credentials = Credentials.read(
            credentials_file_path=credentials_file_path)

        return json.dumps(
            {
                'credentials_file_path': credentials.credentials_file_path,
                'data': credentials.data
            }
        )

    def update(self, credentials_file_path, **kwargs):
        credentials = Credentials.read(
            credentials_file_path=credentials_file_path)
        credentials.update(**kwargs)

        return json.dumps(
            {
                'credentials_file_path': credentials.credentials_file_path,
                'data': credentials.data
            }
        )



import paramiko
from scp import SCPClient
import time
import os
import json


class SSH:
    def __init__(self, ip_address, ssh_key_file_path):
        self.ip_address = ip_address
        self.ssh_key_file_path = ssh_key_file_path

        self.username = 'ec2-user'
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.load_system_host_keys()

    def connect(self, retries=0):
        if retries > 3:
            return False
        interval = 5
        try:
            retries += 1
            self.client.connect(
                hostname=self.ip_address,
                username=self.username,
                key_filename=self.ssh_key_file_path
            )
            return True
        except Exception as e:
            print(e)
            time.sleep(interval)
            print('Retrying SSH connection to {}'.format(self.ip_address))
            self.connect(retries)

    def run(self, cmd):
        try:
            stdin, stdout, stderr = self.client.exec_command(cmd)
            output = stdout.read().decode()
            error = stderr.read().decode()
            return output, error
        except Exception as Error:
            raise Exception(Error.args[0])

    def put(self, file):
        try:
            with SCPClient(self.client.get_transport()) as scp:
                scp.put(file)
            return True
        except Exception as Error:
            raise Exception(Error.args[0])

    def get(self, file, local_path):
        try:
            with SCPClient(self.client.get_transport()) as scp:
                scp.get(file, local_path)

            return True
        except Exception as Error:
            raise Exception(Error.args[0])

    def close(self):
        try:
            self.client.close()
            return True
        except Exception as e:
            print(e)
            return False


def install_packages(ip_address, ssh_key_file_path):
    try:
        print('Installing Packages')
        ssh = SSH(ip_address=ip_address, ssh_key_file_path=ssh_key_file_path)
        ssh.connect()

        output, error = ssh.run('sudo yum update -y')

        ssh.put('app/utilities/ssh/install.sh')
        output, error = ssh.run('ls')
        assert bool(error) is False, error
        home_dir = output.split('\n')
        assert 'install.sh' in home_dir, 'install.sh was not uploaded'

        output, error = ssh.run('bash install.sh')

        output, error = ssh.run(
            'sudo yum list installed | grep -i cloudhsm-client')
        yum_list_grep = output.split()
        assert bool(yum_list_grep), 'cloudhsm-client was not installed.'

        output, error = ssh.run('rm cloudhsm-client-latest.el7.x86_64.rpm')
        assert bool(error) is False, error
        output, error = ssh.run('rm install.sh')
        assert bool(error) is False, error

        ssh.close()
        return
    except Exception as Error:
        raise Exception(Error.args[0])


def upload_file_to_instance(ip_address, ssh_key_file_path, file_path):
    try:
        ssh = SSH(ip_address=ip_address, ssh_key_file_path=ssh_key_file_path)
        ssh.connect()

        ssh.put(file_path)
        output, error = ssh.run('ls')
        assert bool(error) is False, error
        home_dir = output.split('\n')
        assert file_path.split(
            '/')[-1] in home_dir, f"{file_path.split('/')[-1]} was not uploaded."
        return

    except Exception as Error:
        raise Exception(Error.args[0])


def activate_cluster(
        ip_address, ssh_key_file_path, eni_ip, crypto_officer_username, crypto_officer_password, crypto_user_username, crypto_user_password):
    try:
        ssh = SSH(ip_address=ip_address, ssh_key_file_path=ssh_key_file_path)
        ssh.connect()
        output, error = ssh.run(
            f'script activate -eniip {eni_ip} -cousername {crypto_officer_username} -copassword {crypto_officer_password} -cuusername {crypto_user_username} -cupassword {crypto_user_password}')
        assert bool(error) is False, error
        return output

    except Exception as Error:
        raise Exception(Error.args[0])


def configure_cloudhsm_client(ip_address, ssh_key_file_path, hsm_ip_address):
    print('Configure CloudHSM-Client')
    try:
        ssh = SSH(ip_address=ip_address, ssh_key_file_path=ssh_key_file_path)
        ssh.connect()

        output, error = ssh.run(
            f'sudo /opt/cloudhsm/bin/configure -a {hsm_ip_address}')
        assert bool(error) is False, error

        ssh.get('/opt/cloudhsm/etc/cloudhsm_client.cfg')
        with open('cloudhsm_client.cfg') as file:
            data = file.read()
        cloudhsm_configure = json.loads(data)
        os.remove('cloudhsm_client.cfg')
        assert cloudhsm_configure['server']['hostname'] == hsm_ip_address, 'CloudHSM Client Configue Failed'

        ssh.close()
        return
    except Exception as Error:
        raise Exception(Error.args[0])


def gen_ecc_key_pair(ip_address, ssh_key_file_path, eni_ip, crypto_user_username, crypto_user_password, key_label):
    try:
        ssh = SSH(ip_address=ip_address, ssh_key_file_path=ssh_key_file_path)
        ssh.connect()
        output, error = ssh.run(
            f'script gen-ecc-key-pair -eniip {eni_ip} -username {crypto_user_username} -password {crypto_user_password} -label {key_label}')
        assert bool(error) is False, error
        ssh.close()

        return json.loads(output)

    except Exception as Error:
        raise Exception(Error.args[0])


def download_file_from_instance(ip_address, ssh_key_file_path, file, local_path):
    ssh = SSH(ip_address=ip_address, ssh_key_file_path=ssh_key_file_path)
    ssh.connect()
    resp = ssh.get(file=file, local_path=local_path)
    ssh.close()
    return


def sign_tx(ip_address, ssh_key_file_path, eni_ip, unsigned_tx_file, pub_key_handle, private_key_handle, crypto_user_username, crypto_user_password, count, cluster_path):

    ssh = SSH(ip_address=ip_address, ssh_key_file_path=ssh_key_file_path)
    ssh.connect()
    ssh.put(unsigned_tx_file['file_path'])

    cmds = [
        'script',
        'sign',
        '-eniip', eni_ip,
        '-username', crypto_user_username,
        '-password', crypto_user_password,
        '-tx', unsigned_tx_file['file_name'],
        '-vkhandle', pub_key_handle,
        '-skhandle', private_key_handle,
        '-count', str(count)
    ]
    output, error = ssh.run(' '.join(cmds))
    assert bool(error) is False, error

    signature_file_name = output.split()[0]
    assert signature_file_name == f"signedTx{count}.der"

    ssh.get(signature_file_name, cluster_path)

    output, error = ssh.run(f"rm {unsigned_tx_file['file_name']}")
    assert bool(error) is False, error

    output, error = ssh.run(f"rm {signature_file_name}")
    assert bool(error) is False, error

    ssh.close()

    return os.path.join(cluster_path, signature_file_name)



from app.models.ssh_key_model import SSHKey
from app.models.instance_model import Instance
from app.models.cluster_model import Cluster
from app.models.hsm_model import HSM
from app.models.certificate_model import Certs
from app.controllers.credentials_controller import CredentialsController
import app.utilities.ssh as ssh
from app.utilities.terraform import Tf
import os
import time


class Setup:
    def __init__(self, ec2, cloudhsmv2, resource, base_path, aws_region, aws_access_key_id, aws_secret_access_key,
                 customer_ca_key_password, crypto_officer_password, crypto_user_username, crypto_user_password):
        _check_packages(packages=['aws', 'terraform'])

        self.ec2 = ec2
        self.cloudhsmv2 = cloudhsmv2
        self.resource = resource
        self.base_path = base_path
        self.aws_region = aws_region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.customer_ca_key_password = customer_ca_key_password
        self.crypto_officer_password = crypto_officer_password
        self.crypto_user_username = crypto_user_username
        self.crypto_user_password = crypto_user_password

    def run(self):
        ssh_key = _get_ssh_key(client=self.ec2)

        resp = _build_infrastructure(
            ssh_key_name=ssh_key.name,
            region=self.aws_region,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key
        )

        cluster = _cluster(id=resp['cluster_id'], client=self.cloudhsmv2)

        cluster_path = _set_cluster_path(
            base_path=self.base_path, cluster=cluster)

        ssh_key_file_path = _write_ssh_key_to_file(
            ssh_key=ssh_key, cluster_path=cluster_path)

        instance = _instance(
            resource=self.resource, id=resp['instance_id'], ssh_key=ssh_key)

        resp = _create_credentials(
            base_path=self.base_path,
            cluster_path=cluster_path,
            aws_region=self.aws_region,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            customer_ca_key_password=self.customer_ca_key_password,
            crypto_officer_password=self.crypto_officer_password,
            crypto_user_username=self.crypto_user_username,
            crypto_user_password=self.crypto_user_password,
            cluster_id=cluster.id,
            instance_id=instance.id,
            ssh_key_name=ssh_key.name
        )

        hsm = _hsm(cluster=cluster, client=self.cloudhsmv2)

        certs = _certs(
            cluster=cluster, customer_ca_key_password=self.customer_ca_key_password)

        written = _write_certs_to_file(
            certs=certs, cluster_path=cluster_path, cluster=cluster)

        customer_ca_cert_path = os.path.join(cluster_path, 'customerCA.crt')
        uploaded = _upload_customer_ca_cert(
            instance=instance,
            file_path=customer_ca_cert_path,
            ssh_key=ssh_key
        )

        initiated = _initialize_cluster(cluster=cluster, certs=certs)

        activated = _activate_cluster(
            cluster=cluster,
            instance=instance,
            ssh_key=ssh_key,
            crypto_officer_password=self.crypto_officer_password,
            crypto_user_username=self.crypto_user_username,
            crypto_user_password=self.crypto_user_password
        )

        return {
            'cluster_id': cluster.id,
            'ssh_key_name': ssh_key.name,
            'ssh_key_pem': ssh_key.material,
            'instance_id': instance.id
        }


def _check_packages(packages=[]):
    for package in packages:
        proc = os.system(f'{package} --version')
        if proc != 0:
            raise PackageNotInstalledError(f'{package} is not installed')
    return True


def _get_ssh_key(client):
    return SSHKey.create(client=client)


def _build_infrastructure(region, ssh_key_name, aws_access_key_id, aws_secret_access_key):
    return Tf(
        region=region,
        ssh_key_name=ssh_key_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key
    ).build()


def _cluster(client, id):
    return Cluster(client=client, id=id)


def _set_cluster_path(base_path, cluster):
    cluster_path = os.path.join(base_path, cluster.id)
    if os.path.isdir(cluster_path) is False:
        os.mkdir(cluster_path)
    return cluster_path


def _write_ssh_key_to_file(ssh_key, cluster_path):
    ssh_key_file_path = ssh_key.write_to_file(cluster_path=cluster_path)
    return ssh_key_file_path


def _instance(resource, id, ssh_key):
    instance = Instance(resource=resource, id=id)
    instance.install_packages(ssh_key_file_path=ssh_key.ssh_key_file_path)
    return instance


def _create_credentials(base_path, cluster_path, aws_region, aws_access_key_id, aws_secret_access_key, customer_ca_key_password,
                        crypto_officer_password, crypto_user_username, crypto_user_password, cluster_id, instance_id, ssh_key_name):
    credentials = CredentialsController()
    cedentials = credentials.create(
        base_path=base_path,
        cluster_path=cluster_path,
        aws_region=aws_region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        customer_ca_key_password=customer_ca_key_password,
        crypto_officer_password=crypto_officer_password,
        crypto_user_username=crypto_user_username,
        crypto_user_password=crypto_user_password,
        cluster_id=cluster_id,
        instance_id=instance_id,
        ssh_key_name=ssh_key_name
    )


def _hsm(client, cluster):
    hsm = HSM.create(
        cluster_id=cluster.id,
        availability_zone=cluster.azs[0],
        client=client
    )

    seconds = 0
    while hsm.state != 'ACTIVE':
        time.sleep(10)
        seconds += 10
        elapsed = time.localtime(seconds)
        str_time = time.strftime("%Mm%Ss", elapsed)
        print(
            f'aws_cloudhsm_v2_HSM: Creating ... [{str_time} elapsed]')
    print(f'aws_cloudhsm_v2_HSM: Created! [{str_time} elapsed]')
    return hsm


def _certs(customer_ca_key_password, cluster):
    certs = Certs(
        pem_csr=cluster.csr,
        passphrase=customer_ca_key_password
    )

    return certs


def _write_certs_to_file(certs, cluster_path, cluster):
    # Customer CA Key
    with open(os.path.join(cluster_path, 'customerCA.key'), 'wb') as file:
        file.write(certs.pem_private_key)

    # Customer CA Certificate
    with open(os.path.join(cluster_path, 'customerCA.crt'), 'wb') as file:
        file.write(certs.pem_ca_cert)

    # Cluster CSR
    csr = certs.pem_csr if type(
        certs.pem_csr) is bytes else certs.pem_csr.encode('utf-8')
    with open(os.path.join(cluster_path, f'{cluster.id}_ClusterCSR.csr'), 'wb') as file:
        file.write(csr)

    # Customer HSM Certificate
    with open(os.path.join(cluster_path, f'{cluster.id}_CustomerHSMCertificate.crt'), 'wb') as file:
        file.write(certs.pem_hsm_cert)


def _upload_customer_ca_cert(file_path, instance, ssh_key):
    if os.path.exists(file_path):
        ssh.upload_file_to_instance(
            ip_address=instance.public_ip_address,
            ssh_key_file_path=ssh_key.ssh_key_file_path,
            file_path=file_path
        )
    return True


def _initialize_cluster(cluster, certs):
    cluster.initialize(certs)

    seconds = 0
    while cluster.state != 'INITIALIZED':
        time.sleep(10)
        seconds += 10
        elapsed = time.localtime(seconds)
        str_time = time.strftime("%Mm%Ss", elapsed)
        print(
            f'aws_cloudhsm_v2_Cluster({cluster.id}): Initiating ... [{str_time} elapsed]')
    print(
        f'aws_cloudhsm_v2_Cluster({cluster.id}): Initiated! [{str_time} elapsed]')
    return


def _activate_cluster(cluster, instance, crypto_officer_password, crypto_user_username, crypto_user_password, ssh_key):
    resp = cluster.activate(
        instance=instance,
        crypto_officer_password=crypto_officer_password,
        crypto_officer_username='admin',
        crypto_user_username=crypto_user_username,
        crypto_user_password=crypto_user_password,
        ssh_key=ssh_key
    )

    return resp


class PackageNotInstalledError(Exception):
    pass


class WriteFileError(Exception):
    pass



import uuid
import os


class SSHKey:

    def __init__(self, id, name, material, fingerprint):
        self.id = id
        self.name = name
        self.material = material
        self.fingerprint = fingerprint

    @classmethod
    def create(cls, client):
        name = f'Piggy_SSH_Key_{str(uuid.uuid4())[:8]}'
        resp = client.create_key_pair(KeyName=name)
        key = SSHKey(
            id=resp['KeyPairId'],
            name=resp['KeyName'],
            material=resp['KeyMaterial'],
            fingerprint=resp['KeyFingerprint']
        )
        return key

    @classmethod
    def all(cls, client):
        resp = client.describe_key_pairs()
        key_pairs = resp['KeyPairs']
        for key in key_pairs:
            if key.get('KeyFingerprint') is not None:
                del key['KeyFingerprint']
            if key.get('Tags') is not None:
                del key['Tags']
        return key_pairs

    def write_to_file(self, cluster_path):
        self.ssh_key_file_path = os.path.join(cluster_path, f'{self.name}.pem')
        with open(self.ssh_key_file_path, 'w') as file:
            file.write(self.material)
        return self.ssh_key_file_path



import app.utilities.ssh as ssh
import json


class Cluster:

    def __init__(self, client, id):
        self.client = client
        self.id = id

    @classmethod
    def all(cls, client):
        resp = client.describe_clusters()
        return resp['Clusters']

    @property
    def hsms(self):
        return self.read()['Hsms']

    @property
    def azs(self):
        subnet_mapping = self.read()['SubnetMapping']
        azs = []
        for key, value in subnet_mapping.items():
            azs.append(key)
        return azs

    @property
    def csr(self):
        return self.read()['Certificates']['ClusterCsr']

    @property
    def state(self):
        return self.read()['State']

    def initialize(self, certs):
        assert self.state == 'UNINITIALIZED', 'Cluster state is not UNITIALIZED'
        assert certs.valid, 'Certificates not valid'
        self.client.initialize_cluster(
            ClusterId=self.id,
            SignedCert=certs.pem_hsm_cert.decode('UTF-8'),
            TrustAnchor=certs.pem_ca_cert.decode('UTF-8')
        )
        return

    def activate(
            self, instance, crypto_officer_username, crypto_officer_password, crypto_user_username, crypto_user_password, ssh_key):
        eni_ip = self.hsms[0]['EniIp']
        resp_json = ssh.activate_cluster(
            ip_address=instance.public_ip_address,
            ssh_key_file_path=ssh_key.ssh_key_file_path,
            eni_ip=eni_ip,
            crypto_officer_username=crypto_officer_username,
            crypto_officer_password=crypto_officer_password,
            crypto_user_username=crypto_user_username,
            crypto_user_password=crypto_user_password
        )

        resp = json.loads(resp_json)

        assert resp.get(
            'error') is None, f"Activate cluster error: {resp['error']}"
        assert resp['crypto_officer']['username'] == crypto_officer_username
        assert resp['crypto_officer']['password'] == crypto_officer_password

        return True

    def read(self):
        resp = self.client.describe_clusters(
            Filters={'clusterIds': [self.id]})
        return resp['Clusters'][0]

    def destroy(self):
        return False



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    author="Altmirai LLC",
    author_email='kyle.stewart@altmirai.com',
    python_requires='>=3.9',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.9',
    ],
    description="A command line tool that provides the functionality to use AWS CloudHSM services as a bitcoin wallet.",
    name='piggycli',
    version='0.0.26',
    packages=['app', 'app.models', 'app.utilities', 'app.adapters', 'app.controllers',
              'app.routes', 'app.utilities.terraform', 'app.utilities.bitcoin', 'app.utilities.ssh'],
    install_requires=['click', 'paramiko', 'scp', 'boto3',
                      'base58', 'ecdsa', 'requests', 'blockcypher'],
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': [
            'piggy=app.routes.click:piggy'
        ]
    },
    url='',
    keywords='altpiggybank',
    package_data={
        'app.utilities.terraform': ['*.tf']
    }
)



from datetime import datetime
from app.utilities.bitcoin.addresses import P2PKH
from app.adapters import Explorer

# import hashlib
# from base58 import b58encode_check
# from ecdsa import VerifyingKey
import json
import datetime


class Address():
    def __init__(self, id, pub_key_pem, pub_key_handle, private_key_handle, address, confirmed_balance, spent, txrefs):
        self.id = id
        self.pub_key_pem = pub_key_pem
        self.pub_key_handle = pub_key_handle
        self.private_key_handle = private_key_handle
        self.address = address
        self.confirmed_balance = confirmed_balance
        self.txrefs = txrefs
        self.spent = spent

    @classmethod
    def all(cls, bucket, s3):
        addresses = []
        resp = s3.list_objects(Bucket=bucket)
        keys = resp['Contents']
        for key in keys:
            address = cls.find(bucket=bucket, s3=s3, id=key['Key'])
            addresses.append(address)
        return addresses

    @classmethod
    def find(cls, bucket, s3, id):
        resp = s3.get_object(Bucket=bucket, Key=id)
        data_json = resp['Body'].read().decode()
        data = json.loads(data_json)

        address = cls(
            id=id,
            pub_key_pem=data.get('pub_key_pem'),
            pub_key_handle=data.get('pub_key_handle'),
            private_key_handle=data.get('private_key_handle'),
            address=data.get('address'),
            confirmed_balance=data.get('confirmed_balance'),
            txrefs=data.get('txrefs'),
            spent=data.get('spent')
        )

        return address

    @classmethod
    def create(cls, pub_key):
        address = P2PKH(pem=pub_key.pem).address
        explorer = Explorer(address=address)

        return cls(
            id=pub_key.label,
            pub_key_pem=pub_key.pem,
            pub_key_handle=pub_key.handle,
            private_key_handle=pub_key.private_key_handle,
            address=address,
            confirmed_balance=explorer.confirmed_balance,
            spent=explorer.spent,
            txrefs=explorer.txrefs
        )

    def save(self, bucket, s3, region):
        if bucket_exists(bucket=bucket, s3=s3) is False:
            create_bucket(bucket=bucket, s3=s3, region=region)
        key = self.id

        data_json = json.dumps(
            {
                'pub_key_handle': self.pub_key_handle,
                'private_key_handle': self.private_key_handle,
                'pub_key_pem': self.pub_key_pem,
                'address': self.address,
                'confirmed_balance': self.confirmed_balance,
                'spent': self.spent,
                'txrefs': self.txrefs
            }
        )
        data_bytes = bytes(data_json, 'UTF-8')
        resp = s3.put_object(
            Body=data_bytes,
            Bucket=bucket,
            Key=key
        )
        assert resp['ResponseMetadata'][
            'HTTPStatusCode'] == 200, f'Failed to save address: {self.id} to bucket: {bucket}'
        return self.id

    def refresh(self, bucket, s3, region):
        explorer = Explorer(address=self.address)
        self.confirmed_balance = explorer.confirmed_balance
        self.spent = explorer.spent
        self.txrefs = explorer.txrefs
        self.save(bucket=bucket, s3=s3, region=region)
        return self


def bucket_exists(bucket, s3):
    resp = s3.list_buckets()
    for _bucket in resp['Buckets']:
        if _bucket['Name'] == bucket:
            return True
    return False


def create_bucket(bucket, s3, region):
    location = {'LocationConstraint': region}
    resp = s3.create_bucket(
        Bucket=bucket,
        CreateBucketConfiguration=location
    )

    return resp['Location']



from app.controllers.setup_controller import Setup
from app.controllers.credentials_controller import CredentialsController
from app.controllers.status_controller import StatusController
from app.controllers.addresses_controller import AddressController
from app.controllers.tx_controller import TxController
from app.utilities.decorators import creds, check_status

import boto3
import click


class NotRequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if = kwargs.pop('not_required_if')
        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs['help'] = (kwargs.get('help', '') +
                          ' NOTE: This argument is mutually exclusive with %s' %
                          self.not_required_if
                          ).strip()
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        not_required_true = ctx.params.get(self.not_required_if)

        if not_required_true:
            if we_are_present:
                raise click.UsageError(
                    "Illegal usage: `%s` is mutually exclusive with `%s`" % (
                        self.name, self.not_required_if))
            else:
                self.prompt = None
                self.required = False

        return super(NotRequiredIf, self).handle_parse_result(ctx, opts, args)


@click.group()
def piggy():
    pass


@piggy.command()
@click.option('-path', 'path', prompt='Path', required=True)
@click.option('-region', 'aws_region', prompt='AWS Region', required=True)
@click.option('-id', 'aws_access_key_id', prompt='AWS Access Key ID', required=True)
@click.option('-key', 'aws_secret_access_key', prompt='AWS Secret Access Key', required=True)
@click.option('-customer_ca_key_password', 'customer_ca_key_password', prompt='Customer CA Key Password', required=True)
@click.option('-crypto_officer_password', 'crypto_officer_password', prompt='Crypto Officer Password', required=True)
@click.option('-crypto_user_username', 'crypto_user_username', prompt='Crypto User Username', required=True)
@click.option('-crypto_user_password', 'crypto_user_password', prompt='Crypto User Password', required=True)
def setup(**kwargs):
    aws_access_key_id = kwargs['aws_access_key_id']
    aws_secret_access_key = kwargs['aws_secret_access_key']

    ec2 = boto3.client('ec2', aws_access_key_id=aws_access_key_id,
                       aws_secret_access_key=aws_secret_access_key)
    cloudhsmv2 = boto3.client(
        'cloudhsmv2', aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key)
    resource = boto3.resource(
        'ec2', aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key)

    setup = Setup(
        ec2=ec2,
        cloudhsmv2=cloudhsmv2,
        resource=resource,
        base_path=kwargs['path'],
        aws_region=kwargs['aws_region'],
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        customer_ca_key_password=kwargs['customer_ca_key_password'],
        crypto_officer_password=kwargs['crypto_officer_password'],
        crypto_user_username=kwargs['crypto_user_username'],
        crypto_user_password=kwargs['crypto_user_password']
    )

    resp = setup.run()
    click.echo()
    click.secho(resp, fg='green')


@piggy.group()
def credentials():
    pass


@credentials.command()
@click.option('-path', 'path', type=click.Path(), prompt='Path', required=True)
@click.option('-region', 'aws_region', prompt='AWS Region', required=True)
@click.option('-ssh_key_name', 'ssh_key_name', prompt='SSH Key Name', required=True)
@click.option('-cluster_id', 'cluster_id', prompt='Cluster ID', required=True)
@click.option('-instance_id', 'instance_id', prompt='Instance ID', required=True)
@click.option('-aws_access_key_id', 'aws_access_key_id', prompt='AWS Access Key ID', required=True)
@click.option('-aws_secret_access_key', 'aws_secret_access_key', prompt='AWS Secret Access Key', required=True)
@click.option('-customer_ca_key_password', 'customer_ca_key_password', prompt='Customer CA Key Password', required=True)
@click.option('-crypto_officer_password', 'crypto_officer_password', prompt='Crypto Officer Password', required=True)
@click.option('-crypto_user_username', 'crypto_user_username', prompt='Crypto User Username', required=True)
@click.option('-crypto_user_password', 'crypto_user_password', prompt='Crypto User Password', required=True)
def create(**kwargs):
    credentials = CredentialsController()
    resp = credentials.create(**kwargs)
    click.echo()
    click.secho(resp, fg='green')
    click.echo()


@credentials.command()
@click.option('-region', 'aws_region', required=False)
@click.option('-ssh_key_name', 'ssh_key_name', required=False)
@click.option('-cluster_id', 'cluster_id', required=False)
@click.option('-instance_id', 'instance_id', required=False)
@click.option('-aws_access_key_id', 'aws_access_key_id', required=False)
@click.option('-aws_secret_access_key', 'aws_secret_access_key', required=False)
@click.option('-customer_ca_key_password', 'customer_ca_key_password', required=False)
@click.option('-crypto_officer_password', 'crypto_officer_password', required=False)
@click.option('-crypto_user_username', 'crypto_user_username', required=False)
@click.option('-crypto_user_password', 'crypto_user_password', required=False)
@click.option('-creds', 'credentials_file_path', type=click.Path(), required=False)
@creds
def update(credentials, **kwargs):
    update_dict = {}
    for key, value in kwargs.items():
        if bool(value):
            update_dict[key] = value

    resp = credentials.update(
        credentials_file_path=credentials.credentials_file_path, **update_dict)

    click.echo()
    click.secho(resp.data, fg='green')
    click.echo()


@piggy.command()
@click.option('-creds', 'credentials_file_path', type=click.Path(), required=False)
@click.option('-sleep', 'action', flag_value='sleep', default=False)
@click.option('-wake', 'action', flag_value='wake', default=False)
@creds
def status(credentials, action):
    status = StatusController(credentials=credentials)

    if action == 'wake':
        click.echo()
        if click.confirm(click.style('Are you sure you want to wake the pig, starting an HSM costs money?', fg='red')):
            resp = status.wake()
            click.echo()
            click.secho(resp, fg='green')
            click.echo()
    elif action == 'sleep':
        click.echo()
        if click.confirm(click.style('Are you sure you want to put the pig to sleep?', fg='red')):
            resp = status.sleep()
            click.echo()
            click.secho(resp, fg='green')
            click.echo()
    else:
        resp = status.show()
        click.echo()
        click.secho(resp, fg='green')
        click.echo()


@piggy.group()
def address():
    pass


@address.command()
@click.option('-creds', 'credentials_file_path', type=click.Path(), required=False)
@creds
def list(credentials):
    controller = AddressController(credentials=credentials)
    resp = controller.index()
    click.echo('')
    for address in resp['data']['addresses']:
        if address.spent is True:
            color = 'red'
        elif address.confirmed_balance > 0:
            color = 'green'
        else:
            color = 'blue'

        click.secho(
            f"id: {address.id}, address: {address.address}, confirmed_balance: {address.confirmed_balance}, spent: {address.spent}", fg=color)
    click.echo()


@address.command()
@click.option('-creds', 'credentials_file_path', type=click.Path(), required=False)
@creds
@check_status
def create(credentials):
    controller = AddressController(credentials=credentials)
    resp = controller.create()
    address = resp['data']['address']
    click.echo()
    click.secho(f'id: {address.id}', fg='green')
    click.secho(f'address: {address.address}', fg='green')
    click.secho(
        f'confirmed_balance: {address.confirmed_balance}', fg='green')
    click.secho(f'spent: {address.spent}', fg='green')
    click.echo()
    click.secho(f'public_key_handle: {address.pub_key_handle}', fg='green')
    click.secho(
        f'private_key_handle: {address.private_key_handle}', fg='green')
    click.echo()
    click.secho('public_key_pem: ', fg='green')
    click.secho(address.pub_key_pem, fg='green')
    click.echo()


@address.command()
@click.option('-creds', 'credentials_file_path', type=click.Path(), required=False)
@click.option('-id', 'id', prompt='Address ID', required=True)
@creds
def show(credentials, id):
    controller = AddressController(credentials=credentials)
    resp = controller.show(id=id)
    address = resp['data']['address']

    click.echo()
    click.secho(f'id: {address.id}', fg='green')
    click.secho(f'address: {address.address}', fg='green')
    click.secho(
        f'confirmed_balance: {address.confirmed_balance}', fg='green')
    click.secho(f'spent: {address.spent}', fg='green')
    click.echo()
    click.secho(f'public_key_handle: {address.pub_key_handle}', fg='green')
    click.secho(
        f'private_key_handle: {address.private_key_handle}', fg='green')
    click.echo()
    click.secho('public_key_pem: ', fg='green')
    click.secho(address.pub_key_pem, fg='green')
    click.echo()
    click.secho('txrefs:', fg='green')
    click.secho(address.txrefs, fg='green')
    click.echo()


@address.command()
@click.option('-id', 'id', prompt='Address ID', required=True)
@click.option('-creds', 'credentials_file_path', type=click.Path(), required=False)
@creds
def update(credentials, id):
    controller = AddressController(credentials=credentials)
    resp = controller.update(id=id)
    address = resp['data']['address']

    click.echo()
    click.secho(f'id: {address.id}', fg='green')
    click.secho(f'address: {address.address}', fg='green')
    click.secho(
        f'confirmed_balance: {address.confirmed_balance}', fg='green')
    click.secho(f'spent: {address.spent}', fg='green')
    click.echo()
    click.secho(f'public_key_handle: {address.pub_key_handle}', fg='green')
    click.secho(
        f'private_key_handle: {address.private_key_handle}', fg='green')
    click.echo()
    click.secho('public_key_pem: ', fg='green')
    click.secho(address.pub_key_pem, fg='green')
    click.echo()
    click.secho('txrefs:', fg='green')
    click.secho(address.txrefs, fg='green')
    click.echo()


@piggy.command()
@click.option('-all', 'all', is_flag=True, required=True, prompt="Send recipient all the BTC in address", cls=NotRequiredIf, not_required_if='partial')
@click.option('-some', 'partial', is_flag=True)
@click.option('-from', 'address_id', prompt='Sending Address ID', required=True)
@click.option('-to', 'recipient', prompt='Recipient Addreess', required=True)
@click.option('-fee', 'fee', type=click.INT, prompt='Mining Fee', required=True)
@click.option('-qty', 'value', type=click.INT, prompt='Quantity to send',  cls=NotRequiredIf, not_required_if='all')
@click.option('-change', 'change_address', required=True, prompt='Change address',
              cls=NotRequiredIf, not_required_if='all')
@click.option('-creds', 'credentials_file_path', type=click.Path(), required=False)
@creds
@check_status
def send(credentials, address_id, recipient, all, partial, fee, value, change_address):
    controller = TxController(credentials=credentials)
    valid = controller.validate(address_id=address_id, recipient=recipient,
                                all=all, fee=fee, value=value, change_address=change_address)

    if valid.get('error') is not None:
        click.echo()
        click.secho(f"Danger Will Robinson! {valid['error']}", fg="red")
        click.echo()

    elif all:
        click.echo()
        click.secho('Transation Details:', fg='green')
        click.echo()
        click.secho(
            f"Address {valid['address'].address} will send:", fg='green')
        click.secho(
            f"  * {valid['value']} SATs to {valid['recipient']}, and", fg='green')
        click.secho(f"  * pay a {fee} SATs mining fee.", fg='green')
        click.echo()

        if click.confirm('Confirm details'):
            tx_hex = controller.create(**valid)

            click.echo()
            click.secho('Raw Tx Hex:', fg='green')
            click.echo()
            click.secho(tx_hex, fg='green')
            click.echo()
    else:
        click.echo('')
        click.echo('Transation Details:')
        click.echo()
        click.echo(f"Address {valid['address'].address} will sends:")
        click.echo(
            f"  * {valid['value']} SATs to {valid['recipient']},")
        click.echo(
            f"  * {valid['change']} SATs to {valid['change_address']}, and")
        click.echo(f"  * pay a {fee} SATs mining fee.")
        click.echo('')

        if click.confirm('Confirm details'):
            tx_hex = controller.create(**valid)

            click.echo()
            click.echo('Raw Tx Hex:')
            click.echo(tx_hex)
            click.echo()



import subprocess
import json
from pathlib import Path


class Tf:
    def __init__(self, region, ssh_key_name, aws_access_key_id, aws_secret_access_key):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.ssh_key_name = ssh_key_name
        self.dir = Path(__file__).parent

    @property
    def cluster_id(self):
        return self.outputs['cluster_id']['value'] if self.outputs else None

    @property
    def vpc_id(self):
        return self.outputs['vpc_id']['value'] if self.outputs else None

    @property
    def ec2_instance_id(self):
        return self.outputs['ec2_instance_id']['value'] if self.outputs else None

    @property
    def outputs(self):
        with open('app/utilities/terraform/terraform.tfstate', 'r') as file:
            state = json.load(file)
        return state['outputs']

    def init(self):
        cmds = ['terraform', 'init']
        resp = subprocess.run(cmds, cwd=self.dir, capture_output=False)
        assert resp.returncode == 0, 'Terraform initialization failed'
        return True

    def init_validate(self):
        cmds = ['terraform', 'init', '-var', 'backend=false']
        resp = subprocess.run(cmds, cwd=self.dir, capture_output=False)
        assert resp.returncode == 0, 'Terraform initialization failed'
        return True

    def validate(self):
        self.init_validate()
        cmds = ['terraform', 'validate']
        resp = subprocess.run(cmds, cwd=self.dir, capture_output=False)
        assert resp.returncode == 0, 'Terraform validation failed'
        self._clean_up()
        return True

    def build(self):
        self.init()
        cmds = self._add_vars(['terraform', 'apply', '-auto-approve'])
        resp = subprocess.run(cmds, cwd=self.dir, capture_output=False)
        assert resp.returncode == 0, 'Terraform build failed'
        self._clean_up()
        return {'cluster_id': self.cluster_id, 'vpc_id': self.vpc_id, 'instance_id': self.ec2_instance_id}

    def destroy(self):
        self.init()
        cmds = self._add_vars(['terraform', 'destroy', '-auto-approve'])
        resp = subprocess.run(cmds, cwd=self.dir, capture_output=False)
        assert resp.returncode == 0, 'Terraform destroy failed'
        self._clean_up()
        return True

    def _add_vars(self, cmds):
        vars = {"region": self.region, "ssh_key_name": self.ssh_key_name,
                "aws_access_key_id": self.aws_access_key_id, 'aws_secret_access_key': self.aws_secret_access_key}
        for var, value in vars.items():
            cmds.append('-var')
            cmds.append(f'{var}={value}')
        return cmds

    def _clean_up(self):
        cmds = ['rm', '-r', '.terraform']
        resp = subprocess.run(cmds, cwd=self.dir, capture_output=False)
        assert resp.returncode == 0, 'Terraform clean up failed'
        return True
